from datetime import timedelta
from typing import Tuple
import pytest
import pytest_mock
from src.rlretry.rlretry import (
    RLAgent,
    RLEnvironment,
    Action,
    rlretry,
    RLRetryError,
    default_alpha_func,
)
import random

ACTION_DURATIONS = {
    Action.ABRT: timedelta(seconds=0.0),
    Action.RETRY0: timedelta(seconds=1.0),
    Action.RETRY0_1: timedelta(seconds=1.1),
    Action.RETRY0_2: timedelta(seconds=1.2),
    Action.RETRY0_5: timedelta(seconds=1.5),
}


@pytest.mark.parametrize("retries", [2, 3, 4])
@pytest.mark.parametrize("iterations", [10, 20, 50])  # [10, 20, 50, 100])
@pytest.mark.parametrize("timeout_int", [1, 2, 5, 10, 20])  # [1, 2, 5, 10, 20])
@pytest.mark.parametrize("fail_duration_int", [0.02, 0.1, 1])  # [0.02, 0.1, 1])
def test_scenario_1(mocker, retries, timeout_int, iterations, fail_duration_int):
    """
    this scenario tests that if we are faced with a state that always results in failure
    then ABRT ends up with the highest average reward
    """

    random.seed(0)

    def dummy_func():
        raise RuntimeError("badness ocurred")

    def mock_execute(enviro: RLEnvironment, action: Action) -> Tuple[str, float]:
        if action == Action.ABRT:
            # need to give -1 reward to the state/action
            next_state = "abort"
            duration = timedelta(seconds=0.01)
        else:
            next_state = enviro.run_func()
            duration = (
                timedelta(seconds=fail_duration_int)
                + enviro._max_wait * action.sleeptime()
            )

        reward = enviro.next_state_to_reward(next_state, duration)

        # print(f"mock_execute({next_state}, {action}) -> {next_state, reward}")

        return next_state, reward

    mocker.patch("src.rlretry.rlretry.RLEnvironment.execute_action", mock_execute)

    timeout = timedelta(seconds=timeout_int)

    wrapped_func, agent = rlretry(
        timeout=timeout,
        max_retries=retries,
        raise_primary_exception=False,
        alpha=default_alpha_func,
        optimistic_initial_values=True,
    )(dummy_func, return_agent=True)

    for _ in range(iterations):
        try:
            wrapped_func()
        except RLRetryError:
            pass

    # print()
    # print(agent._state_action_map._df)
    # print(agent._state_action_map._counts_df)
    assert agent._state_action_map.best_action("RuntimeError") == Action.ABRT


@pytest.mark.parametrize("retries", [2, 3, 4])
@pytest.mark.parametrize("timeout_int", [1, 2, 5, 10, 20])
@pytest.mark.parametrize("iterations", [10, 20, 50, 100])
def test_scenario_2(mocker, retries, timeout_int, iterations):
    """
    test to make sure that aborting doesn't result in a higher average reward than
    waiting for ages and then getting success.
    """

    def dummy_func():
        raise RuntimeError("badness ocurred")

    def mock_execute(enviro: RLEnvironment, action: Action) -> Tuple[str, float]:
        if action == Action.ABRT:
            # need to give -1 reward to the state/action
            next_state = "abort"
            duration = timedelta(seconds=0.01)
        else:
            next_state = enviro.run_func()
            if action == Action.RETRY0_5:
                next_state = "success"
            duration = timedelta(seconds=action.sleeptime() + 0.01)

        reward = enviro.next_state_to_reward(next_state, duration)

        # print(f"mock_execute({next_state}, {action}) -> {next_state, reward}")

        return next_state, reward

    mocker.patch("src.rlretry.rlretry.RLEnvironment.execute_action", mock_execute)

    timeout = timedelta(seconds=timeout_int)

    wrapped_func, agent = rlretry(
        timeout=timeout,
        max_retries=retries,
        raise_primary_exception=False,
        alpha=default_alpha_func,
        optimistic_initial_values=True,
    )(dummy_func, return_agent=True)

    for _ in range(iterations):
        try:
            wrapped_func()
        except RLRetryError:
            pass

    # print()
    # print(agent._state_action_map._df)
    # print(agent._state_action_map._counts_df)
    assert agent._state_action_map.best_action("RuntimeError") == Action.RETRY0_5



from __future__ import annotations
from datetime import datetime, timedelta
import random
import time
from typing import Callable, Tuple, Union
from enum import Enum
import pandas as pd
import logging

log = logging.Logger(__name__)


def update_average(
    q0: pd.DataFrame,
    n0: pd.DataFrame,
    q01: pd.DataFrame,
    n01: pd.DataFrame,
    q02: pd.DataFrame,
    n02: pd.DataFrame,
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """
    calculate the total average from two different estimates of the average"""
    denom = n01 + n02 - n0
    return q01 * n01 / denom + q02 * n02 / denom - q0 * n0 / denom, denom


def update_recency_weighted_average(
    q0: pd.DataFrame,
    n0: pd.DataFrame,
    q01: pd.DataFrame,
    n01: pd.DataFrame,
    q02: pd.DataFrame,
    n02: pd.DataFrame,
    alpha: float,
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    # calculate the average reward between q02 and q0 and then replay this average reward onto q01
    k = n02 - n0

    counts = n01 + n02 - n0
    new_reward = (1 - alpha) ** k * q01 + (q02 - q0 * (1 - alpha) ** k)

    return new_reward, counts


class Action(Enum):
    ABRT = 0
    RETRY0 = 1
    RETRY0_1 = 2
    RETRY0_2 = 3
    RETRY0_5 = 4

    def sleeptime(self) -> float:
        if self == Action.RETRY0_1:
            return 0.1
        elif self == Action.RETRY0_2:
            return 0.2
        elif self == Action.RETRY0_5:
            return 0.5
        return 0


class RLRetryError(RuntimeError):
    def __init__(self, e: Exception, msg: str = ""):
        super().__init__(msg)
        self.original_exception = e


class RLRetryTimeout(RLRetryError):
    pass


class RLRetryMaxRetries(RLRetryError):
    pass


class RLRetryAbort(RLRetryError):
    pass


class RLRetryNoException(RuntimeError):
    pass


def default_state_func(e: Exception) -> str:
    return e.__class__.__name__

def default_alpha_func(n: int) -> float:
    '''
    n is the count of attempts made so far for this state/action pair
    '''
    if n > 5:
        return 0.05
    return 0.5 - n * 0.45 / 6


class StateActionMap:
    def __init__(
        self,
        df: pd.DataFrame,
        counts_df: pd.DataFrame,
        initial_value: float = 0.0,
        alpha: Union[float, None, Callable[[int], float]] = 0.0,
    ):
        self._df = StateActionMap.default_df() if df is None or df.empty else df
        self._counts_df = (
            StateActionMap.default_counts_df()
            if counts_df is None or counts_df.empty
            else counts_df
        )
        self._last_saved_df = self._df.copy(deep=True)
        self._last_saved_counts_df = self._counts_df.copy(deep=True)
        self._initial_value = initial_value

        if callable(alpha):
            self._alpha = alpha
        elif isinstance(alpha, float):
            self._alpha = lambda _: alpha
        else:
            self._alpha = None


    @staticmethod
    def default_df() -> pd.DataFrame:
        return pd.DataFrame(columns=list(Action), dtype=pd.Float32Dtype())

    @staticmethod
    def default_counts_df() -> pd.DataFrame:
        return pd.DataFrame(columns=list(Action))

    @staticmethod
    def random_action() -> Action:
        return random.choice(list(Action))

    def best_action(self, state) -> Action:
        if state not in self._df.index:
            return random.choice(list(Action))

        actions = self._df.loc[state]

        idx_of_best_action = actions.argmax()

        best_action = actions.index[idx_of_best_action]

        return best_action

    def create_state(self, state: str):
        self._df.loc[state] = [float(self._initial_value) for _ in list(Action)]
        self._counts_df.loc[state] = [0 for _ in list(Action)]

    def update_average_reward(self, state: str, action: Action, new_reward: float):
        if state not in self._df.index:
            self.create_state(state)
        count = self._counts_df.loc[state][action]
        current_average = self._df.loc[state][action]
        value_delta = new_reward - current_average
        # if alpha has been specified, use that as a recency weighting
        # otherwise use average reward
        if self._alpha is not None:
            value_delta *= self._alpha(count)
        else:
            value_delta /= count + 1

        self._df.loc[state][action] += value_delta
        self._counts_df.loc[state][action] += 1

    def update_last_saved(self):
        self._last_saved_df = self._df.copy(deep=True)
        self._last_saved_counts_df = self._counts_df.copy(deep=True)


def _default_weight_loader() -> Tuple[pd.DataFrame, pd.DataFrame]:
    return (StateActionMap.default_df(), StateActionMap.default_counts_df())


def _default_weight_dumper(
    _df: pd.DataFrame,
    _counts_df: pd.DataFrame,
    _previous_df: pd.DataFrame,
    _previous_counts_df: pd.DataFrame,
):
    pass


class RLAgent:

    def __init__(
        self,
        epsilon: float,
        weight_loader: Callable[
            [], Tuple[pd.DataFrame, pd.DataFrame]
        ] = _default_weight_loader,
        weight_dumper: Callable[
            [pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame], None
        ] = _default_weight_dumper,
        initial_value: float = 0.0,
        alpha: Union[float, None, Callable[[int], float]] = 0.0,
        dump_interval: int = 100,
    ):
        self._state_action_map = StateActionMap(
            *weight_loader(), initial_value, alpha=alpha
        )
        self._eps = epsilon
        self._age = 0
        self._weight_loader = weight_loader
        self._weight_dumper = weight_dumper
        self._dump_interval = dump_interval

    def dump_weights(self):
        self._weight_dumper(
            self._state_action_map._df,
            self._state_action_map._counts_df,
            self._state_action_map._last_saved_df,
            self._state_action_map._last_saved_counts_df,
        )
        self._state_action_map.update_last_saved()

    def choose_action(self, state: str) -> Action:
        self._age += 1
        if self._age % self._dump_interval == 0:
            self.dump_weights()

        if random.random() < self._eps:
            log.debug("agent choosing random action")
            return self._state_action_map.random_action()
        return self._state_action_map.best_action(state)

    def apply_reward(self, state, action, reward):
        self._state_action_map.update_average_reward(state, action, reward)


class RLEnvironment:

    def __init__(
        self,
        func,
        max_wait: timedelta,
        state_func: Callable[[Exception], str] = default_state_func,
    ):
        self._previous_action_start_time = datetime.utcnow()
        self._func = func
        self._state_func = state_func
        self.func_retval = None
        self._max_wait = max_wait
        self.last_exception = RLRetryNoException("something has gone wrong")

    def run_func(self) -> str:
        try:
            self.func_retval = self._func()
            return "success"
        except Exception as e:
            self.last_exception = e
            return self._state_func(e)

    def next_state_to_reward(self, next_state: str, duration: timedelta) -> float:
        if next_state == "success":
            return 2.5 - duration / self._max_wait
        return 1 - duration / self._max_wait

    def execute_action(self, action: Action) -> Tuple[str, float]:
        log.debug(f"RLEnvironment execute_action({action})")
        previous_action_start_time = datetime.utcnow()

        next_state = self.run_func() if action != Action.ABRT else "abort"

        time.sleep(self._max_wait.total_seconds() * action.sleeptime())

        reward = self.next_state_to_reward(
            next_state, datetime.utcnow() - previous_action_start_time
        )

        return next_state, reward

def rlretry(
    max_retries: int = 5,
    timeout: timedelta = timedelta(seconds=300),
    state_func: Callable = default_state_func,
    epsilon: float = 0.1,
    weight_loader: Callable[
        [], Tuple[pd.DataFrame, pd.DataFrame]
    ] = _default_weight_loader,
    weight_dumper: Callable[
        [pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame], None
    ] = _default_weight_dumper,
    dump_interval: int = 100,
    optimistic_initial_values: bool = True,
    alpha: Union[float, None, Callable[[int], float]] = default_alpha_func,
    raise_primary_exception=False,
):
    """
    A decorator that performs retries on a function if it throws an exceptions.

    It uses reinforcement learning to determine the best course of action to take on any given exception
    It will either abort, retry immediately or retry after one of 4 time periods (which are scaled relative to timeout)


    :param max_retries: the maximum number of retries to perform
    :param timeout: the maximum time to allow for all retries
    :param state_func: a function that accepts an exception and returns a string which is the name of a state (in RL parlance).  The default_state_func uses the name of the exception class as the state.
    :param epsilon: determines how much exploration to do.  Set to zero to always use the best option (but slows down learning).  Set to 1 to always use a random option.  Ideally set it somewhere in between
    :param weight_loader loads the weights (ie. the probabilities that certain actions will be chosen)
    :param weight_dumper a function that saves the weights somewhere (so that you don't start from the beginning next time it runs)
    """
    initial_value = 1 if optimistic_initial_values else 0.0

    def raise_rlexception(e: RLRetryError):
        raise e

    def raise_original_exception(e: RLRetryError):
        raise e.original_exception

    raise_exception = (
        raise_original_exception if raise_primary_exception else raise_rlexception
    )

    def decorator_no_args(
        func: Callable, return_agent: bool = False
    ) -> Union[Callable, Tuple[Callable, RLAgent]]:
        agent = RLAgent(
            epsilon, weight_loader, weight_dumper, initial_value, alpha, dump_interval
        )

        def wrapper(*args, **kwargs):
            start_time = datetime.utcnow()
            environment = RLEnvironment(
                lambda: func(*args, **kwargs),
                timeout,
                state_func=state_func,
            )
            current_state = environment.run_func()
            # if it works first time, then we don't have to do any RL stuff
            if current_state == "success":
                return environment.func_retval

            for _ in range(max_retries):
                if datetime.utcnow() - start_time > timeout:
                    raise_exception(RLRetryTimeout(environment.last_exception))
                action = agent.choose_action(current_state)
                previous_state = current_state
                current_state, reward = environment.execute_action(
                    action
                )
                agent.apply_reward(previous_state, action, reward)
                if current_state == "success":
                    return environment.func_retval
                elif current_state == "abort":
                    raise_exception(
                        RLRetryAbort(
                            environment.last_exception,
                            f"encountered a state in which RLRetry thinks it is not worth continuing {previous_state}",
                        )
                    )

            raise_exception(RLRetryMaxRetries(environment.last_exception))

        if return_agent:
            return wrapper, agent
        return wrapper

    return decorator_no_args



from collections import defaultdict
from datetime import datetime, timedelta
import random
import time
from typing import Callable, Dict, Optional


def default_success_func(e: Exception) -> bool:
    return False


def get_optimum_interval(average_rewards: Dict[timedelta, float]) -> timedelta:
    sorted_intervals = sorted(average_rewards.items(), key=lambda x: x[1])
    return sorted_intervals[-1][0]


def get_random_interval(
    time_increment: timedelta, minimum: timedelta, maximum: timedelta
) -> timedelta:
    num_time_increments = random.randint(
        int(minimum / time_increment), int(maximum / time_increment)
    )
    return time_increment * num_time_increments


def get_epsilon_greedy_interval(
    average_rewards: Dict[timedelta, float],
    time_increment: timedelta,
    minimum: timedelta,
    maximum: timedelta,
    epsilon: float,
) -> timedelta:
    if random.random() < epsilon:
        return get_random_interval(time_increment, minimum, maximum)
    return get_optimum_interval(average_rewards)


def auto_request_interval(
    maximum: timedelta,
    minimum: timedelta = timedelta(seconds=0),
    time_increment: timedelta = timedelta(seconds=1),
    success_func: Callable[[Exception], bool] = default_success_func,
    alpha: float = 0.05,
    epsilon: float = 0.05,
):
    """
    A decorator that adjusts the query interval to find the optimum rate to repeat a function.
    seeks to minimize the average interval between successful queries (ie queries which do not raise an exception)
    """

    if minimum % time_increment != timedelta(seconds=0):
        raise ValueError("minimum must be a multiple of time_increment")

    if maximum % time_increment != timedelta(seconds=0):
        raise ValueError("maximum must be a multiple of time_increment")

    if maximum <= minimum:
        raise ValueError("maximum must be greater than minimum")

    def cropping_func(value: timedelta) -> timedelta:
        if value < minimum:
            return minimum
        if maximum is not None and value > maximum:
            return maximum
        return value

    def decorator_no_args(func: Callable):
        average_rewards = defaultdict(float)
        current_interval = minimum
        last_success_time = None
        last_request_time = None
        subsequent_failed_requests = 0
        n = 0

        def wrapper(*args, **kwargs):
            nonlocal last_request_time, last_success_time, average_rewards, alpha, subsequent_failed_requests, time_increment, current_interval, epsilon, n

            last_success_time = (
                datetime.utcnow() if last_success_time is None else last_success_time
            )
            last_request_time = (
                datetime.utcnow() if last_request_time is None else last_request_time
            )

            # wait until it is time to query
            sleep_duration = cropping_func(current_interval) - (
                datetime.utcnow() - last_request_time
            )

            print(f'waiting for {sleep_duration}')

            time.sleep(max(0, sleep_duration.total_seconds()))

            if n % 5 == 0:
                print('average rewards')
                for interval in sorted(average_rewards.keys()):
                    print (f'    {interval}: {average_rewards[interval]}')
            n += 1

            last_request_time = datetime.utcnow()

            is_success = False
            try:
                func(*args, **kwargs)
                is_success = True
            except Exception as e:
                is_success = success_func(e)

            now = datetime.utcnow()
            # update rewards
            if is_success:
                success_interval_seconds = (now - last_success_time).total_seconds()
                reward = -success_interval_seconds
                average_reward_delta = (
                    reward - average_rewards[current_interval]
                ) * alpha
                average_rewards[current_interval] += average_reward_delta
                last_success_time = datetime.utcnow()
                subsequent_failed_requests = 0
            else:
                fail_interval_seconds = (now - last_request_time).total_seconds()
                # this is just an adjustment to the reward so that we punish multiple failures
                average_rewards[current_interval] -= fail_interval_seconds * alpha
                subsequent_failed_requests += 1

            # now choose a new interval if required
            if is_success or subsequent_failed_requests:
                new_interval = get_epsilon_greedy_interval(
                    average_rewards, time_increment, minimum, maximum, epsilon
                )
                # force a change of interval if we had 10 successive failures
                if not is_success and new_interval == current_interval:
                    new_interval = current_interval + time_increment
                    if new_interval > maximum:
                        new_interval = maximum

                current_interval = new_interval
                print(f'updated current_interval to {current_interval}')

        return wrapper

    return decorator_no_args



import pytest
import pandas as pd

from src.rlretry.rlretry import update_average, update_recency_weighted_average

@pytest.mark.parametrize(
        "batch0,batch1,batch2",
        [
            (list(range(9)), list(range(6, 18)), list(range(23, 36, 2))),
            ([], list(range(6, 18)), list(range(23, 36, 2))),
            (list(range(9)), [], list(range(23, 36, 2))),
            (list(range(9)), list(range(6, 18)), []),
            (list(range(9)), [], []),
            ([], list(range(6, 18)), []),
            (list(range(9)), [], []),
        ]
)
def test_3way_average(batch0, batch1, batch2):

    ntotal = len(batch0) + len(batch1) + len(batch2)

    n0 = len(batch0)
    q0 = sum(batch0) / n0 if n0 > 0 else 0
    n01 = len(batch0) + len(batch1)
    n02 = len(batch0) + len(batch2)
    q01 = sum(batch0 + batch1) / n01 if n01 > 0 else 0
    q02 = sum(batch0 + batch2) / n02 if n02 > 0 else 0

    actual_average, actual_count = update_average(q0, n0, q01, n01, q02, n02)

    assert actual_count == ntotal
    assert actual_average == pytest.approx(sum(batch0 + batch1 + batch2) / ntotal)

    q0_df = pd.DataFrame(((q0,q0),(q0, q0)))
    n0_df = pd.DataFrame(((n0,n0), (n0, n0)))
    q01_df = pd.DataFrame(((q01,q01), (q01,q01)))
    n01_df = pd.DataFrame(((n01,n01), (n01,n01)))
    q02_df = pd.DataFrame(((q02,q02),(q02,q02)))
    n02_df = pd.DataFrame(((n02,n02),(n02,n02)))

    actual_average, actual_count = update_average(q0_df, n0_df, q01_df, n01_df, q02_df, n02_df)

    assert actual_average.shape == (2,2)
    assert actual_count.shape == (2,2)

    num_elems = actual_average.size

    assert list(actual_count.values.flatten()) == [ntotal] * num_elems
    assert list(actual_average.values.flatten()) == [pytest.approx(sum(batch0 + batch1 + batch2) / ntotal)] * num_elems


@pytest.mark.parametrize(
        "batch0,batch1,batch2",
        [
            (list(range(9)), list(range(6, 18)), list(range(23, 36, 2))),
            ([], list(range(6, 18)), list(range(23, 36, 2))),
            (list(range(9)), [], list(range(23, 36, 2))),
            (list(range(9)), list(range(6, 18)), []),
            (list(range(9)), [], []),
            ([], list(range(6, 18)), []),
            (list(range(9)), [], []),
        ]
)

def calculate_weighted_average_iter(start_q, r_values, alpha):
    q = start_q
    for r in r_values:
        q += (r - q) * alpha
    return q

def calculate_weighted_average(start_q, r_values, alpha):
    k = len(r_values)
    alpha_series = [r*(1-alpha)**(k-i-1) for i,r in enumerate(r_values)]
    alpha_sum = sum(alpha_series)
    return start_q*(1-alpha)**k + alpha * alpha_sum

def test_calculate_weighted_average():
    # test extreme alpha values
    assert calculate_weighted_average_iter(0.0, [5,5,5,5], 0) == 0
    assert calculate_weighted_average(0, [5,5,5,5], 0) == 0
    assert calculate_weighted_average_iter(0.0, [5,5,5,7], 1) == 7
    assert calculate_weighted_average(0, [5,5,5,7], 1) == 7


    assert calculate_weighted_average_iter(4.0, [4], 0.1) == 4.0
    assert calculate_weighted_average(4.0, [4], 0.1) == 4.0

    assert calculate_weighted_average_iter(4.0, [4,4], 0.1) == 4.0
    assert calculate_weighted_average(4.0, [4,4], 0.1) == 4.0

    assert calculate_weighted_average_iter(4.0, [4,4,4, 4], 0.1) == 4.0
    assert calculate_weighted_average(4.0, [4,4,4,4], 0.1) == 4.0


    assert calculate_weighted_average_iter(5.0, [5,5,5,5], 0.1) == 5.0
    assert calculate_weighted_average(5.0, [5,5,5,5], 0.1) == 5.0

    assert calculate_weighted_average(4.0, [4]*9, 0.1) == pytest.approx(4.0)
    assert calculate_weighted_average_iter(4.0, [4]*9, 0.1) == pytest.approx(4.0)

    assert calculate_weighted_average_iter(0.0, [4]*9, 0.9) == pytest.approx(4.0)
    assert calculate_weighted_average(0.0, [4]*9, 0.9) == pytest.approx(4.0)

@pytest.mark.parametrize('alpha', [(0.1),(0.5),(0.9)])
@pytest.mark.parametrize(
        "batch0,batch1,batch2",
        [
            (list(range(9)), list(range(6, 18)), list(range(23, 36, 2))),
            ([], list(range(6, 18)), list(range(23, 36, 2))),
            (list(range(9)), [], list(range(23, 36, 2))),
            (list(range(9)), list(range(6, 18)), []),
            (list(range(9)), [], []),
            ([], list(range(6, 18)), []),
            (list(range(9)), [], []),
        ]
)
def test_3way_average_weighted(batch0, batch1, batch2, alpha):
    ntotal = len(batch0) + len(batch1) + len(batch2)

    n0 = len(batch0)
    q0 = calculate_weighted_average_iter(0.0, batch0, alpha)
    n01 = len(batch0) + len(batch1)
    n02 = len(batch0) + len(batch2)
    q01 = calculate_weighted_average(q0, batch1, alpha)
    q02 = calculate_weighted_average(q0, batch2, alpha)

    # print(f'test_3way_average_weighted {q0} {n0} | {q01} {n01} | {q02} {n02}')

    actual_average, actual_count = update_recency_weighted_average(q0, n0, q01, n01, q02, n02, alpha)

    assert actual_count == ntotal
    assert actual_average == pytest.approx(calculate_weighted_average(0, batch0 + batch1 + batch2, alpha))

    q0_df = pd.DataFrame(((q0,q0),(q0, q0)))
    n0_df = pd.DataFrame(((n0,n0), (n0, n0)))
    q01_df = pd.DataFrame(((q01,q01), (q01,q01)))
    n01_df = pd.DataFrame(((n01,n01), (n01,n01)))
    q02_df = pd.DataFrame(((q02,q02),(q02,q02)))
    n02_df = pd.DataFrame(((n02,n02),(n02,n02)))

    actual_average, actual_count = update_recency_weighted_average(q0_df, n0_df, q01_df, n01_df, q02_df, n02_df, alpha)

    assert actual_average.shape == (2,2)
    assert actual_count.shape == (2,2)

    num_elems = actual_average.size

    assert list(actual_count.values.flatten()) == [ntotal] * num_elems
    assert list(actual_average.values.flatten()) == [pytest.approx(calculate_weighted_average(0, batch0 + batch1 + batch2, alpha))] * num_elems




from typing import Tuple
import pytest
import pandas as pd

from src.rlretry.rlretry import RLAgent, update_average, update_recency_weighted_average

q_saved = pd.DataFrame([[0]])
n_saved = pd.DataFrame([[0]])


def dummy_weight_dumper(
    q_curr: pd.DataFrame,
    n_curr: pd.DataFrame,
    q_prev: pd.DataFrame,
    n_prev: pd.DataFrame,
):
    global q_saved, n_saved
    if q_saved is None or n_saved is None:
        q_saved = q_curr
        n_saved = n_curr
        return

    q02 = q_saved
    n02 = n_saved

    q_saved, n_saved = update_average(q_prev, n_prev, q02, n02, q_curr, n_curr)


def dummy_weight_dumper_weighted(
    q_curr: pd.DataFrame,
    n_curr: pd.DataFrame,
    q_prev: pd.DataFrame,
    n_prev: pd.DataFrame,
):
    global q_saved, n_saved
    if q_saved is None or n_saved is None:
        q_saved = q_curr
        n_saved = n_curr
        return

    q02 = q_saved
    n02 = n_saved

    q_saved, n_saved = update_recency_weighted_average(
        q_prev, n_prev, q02, n02, q_curr, n_curr, 0.9
    )


def dummy_weight_loader() -> Tuple[pd.DataFrame, pd.DataFrame]:
    return q_saved, n_saved


@pytest.fixture
def agent1():
    return RLAgent(
        0.1, weight_dumper=dummy_weight_dumper, weight_loader=dummy_weight_loader
    )


@pytest.fixture
def agent2():
    return RLAgent(
        0.1, weight_dumper=dummy_weight_dumper, weight_loader=dummy_weight_loader
    )


@pytest.fixture
def agent_weighted():
    return RLAgent(
        0.1,
        weight_dumper=dummy_weight_dumper_weighted,
        weight_loader=dummy_weight_loader,
    )


def test_multi_dump(agent1: RLAgent, agent2: RLAgent):
    global q_saved, n_saved
    data = list(range(21))

    batch0 = data[:7]
    batch1 = data[7:14]
    batch2 = data[14:]
    n0 = len(batch0)
    n1 = len(batch1)
    n2 = len(batch2)

    # agent1 records an average
    q0df = pd.DataFrame([[sum(batch0) / n0]])
    n0df = pd.DataFrame([[n0]])
    sam = agent1._state_action_map
    sam._df = q0df
    sam._counts_df = n0df

    agent1.dump_weights()

    assert n_saved.shape == (1, 1)
    assert q_saved.shape == (1, 1)
    assert n_saved[0][0] == n0
    assert q_saved[0][0] == sum(batch0) / n0

    # now a agent2 counts more averages
    sam2 = agent2._state_action_map
    sam2._df = pd.DataFrame([[sum(batch1) / n1]])
    sam2._counts_df = pd.DataFrame([[n1]])

    agent2.dump_weights()

    assert n_saved.shape == (1, 1)
    assert q_saved.shape == (1, 1)
    assert n_saved[0][0] == n0 + n1
    assert q_saved[0][0] == sum(batch0 + batch1) / (n0 + n1)

    # now agent1 counts different averages on top of the batch from agent2
    sam._df = pd.DataFrame([[sum(batch0 + batch2) / (n0 + n2)]])
    sam._counts_df = pd.DataFrame([[n0 + n2]])

    agent1.dump_weights()

    assert n_saved.shape == (1, 1)
    assert q_saved.shape == (1, 1)
    assert n_saved[0][0] == n0 + n1 + n2
    assert q_saved[0][0] == sum(batch0 + batch1 + batch2) / (n0 + n1 + n2)
